# fingerprintwords.g

FPVERBOSE := false;
Read("listwords.g");

# fingerprintwords
#
# Find fingerprints for a group
#
# G:                   group
# g1, g2:              generators of G
# eltpairtargetlist:   details for the pair (a,b): see below
# tgtorder:            target order for the product ab
# length:              number of fingerprinting words
# checkgens:           check the group <a, b> (not recommended)
#
# eltpairtargetlist is a list of fingerprinting procedures
# Each entry in the list is itself a list, consisting of:
#     word in x,y generating ccl rep for a
#     word in x,y generating ccl rep for b
#     number of fingerprints to find
#     description of the procedure
#
# See the function M24example below for an example of how to
# use the function fingerprintwords().

fingerprintwords := function(G, g1, g2, eltpairtargetlist, tgtorder,
                        length, checkgens)
    
    local fplist, conjlist,
          superfplist, superconjlist,
          warnings, superwarnings,
          counter, pair, words, t, u, target, numfound,
          description, o1, o2, op1, op2;
    
    o1 := Order(g1);
    o2 := Order(g2);

    op1 := Order(MappedWord(eltpairtargetlist[1][1], [x,y], [g1, g2]));
    op2 := Order(MappedWord(eltpairtargetlist[1][2], [x,y], [g1, g2]));

    words := [ ];
    listwords(op1, op2, 10, 1000,
            function(wd, num) Add(words, wd);
            return true; end, true);

    if length > 0 then
        words := List([1..length], x->words[x]);
    fi;
    
    superfplist := [];
    superconjlist := [];
    superwarnings := [];

    for pair in eltpairtargetlist do
        fplist := [];
        conjlist := [];
        warnings := [];
        counter := 0;
    
        t := MappedWord(pair[1], [x,y], [g1, g2]);
        u := MappedWord(pair[2], [x,y], [g1, g2]);
        target := pair[3];  
        description := pair[4];
        
        if FPVERBOSE then
            Print("\n*** ", description, " ***\n");
        fi;

        numfound := 0;
        listwords(o1, o2, 1000, 10^50,
            function(randwd, num)
                local rnd, v, fingerprint, sg, warning, H;
                rnd := MappedWord(randwd, [x, y], [g1, g2]);
                v := u^rnd;
                if Order(t*v) <> tgtorder then
                    return true;
                fi;

                fingerprint := List(words,
                                wd->Order(MappedWord(wd,
                                [x,y], [t,v])));
                if not fingerprint in fplist then
                    numfound := numfound + 1;
                    Add(fplist, fingerprint);
                    Add(conjlist, randwd);
                    if FPVERBOSE then
                        Print("FP no ",
                              numfound, "/", target,
                              " conj by ", randwd, "\n");
                        Print("  (Long) fingerprint is: ", fingerprint, "\n");
                    fi;
                    warning := rec();
                    if checkgens then
                        H := Group(t,v);
                        sg := Size(H);
                        if not sg = Size(G) then
                            warning.sg := sg;
                            warning.sc := Size(Centralizer(G, H));
                            if IsPermGroup(Group(t,v)) then
                                warning.ol := OrbitLengths(H);
                            fi;
                            if FPVERBOSE then
                                Print("     Not a generating triple!\n");
                                Print("     SC = ", warning.sc, "\n");
                                Print("     SG = ", warning.sg, "\n");
                                if IsPermGroup(H) then
                                    Print("     OL = ",
                                          warning.ol, "\n");
                                fi;
                            fi;
                        fi;
                    fi;
                    Add(warnings, warning);
                fi;


                if numfound >= target then
                    return false;
                fi;
                return true;
            end,
            true);

        Add(superfplist, fplist);
        Add(superconjlist, conjlist);
        Add(superwarnings, warnings);
    od;

    return rec(fplist := superfplist, conj := superconjlist,
               warnings := superwarnings,
               words := words,
               descriptions := List(eltpairtargetlist, x->x[4]),
               gen1 := List(eltpairtargetlist, x->x[1]),
               gen2 := List(eltpairtargetlist, x->x[2]));
    
end;

# listdistinguisher
#
# Given a list L of lists of length n, find a subset S of [1..n]
# such that L[1]|S, L[2]|S, ... L[k]|S are all different.
#
# Used for finding "minimal" subsets of fingerprints which can
# still distinguish the different classes of pairs.

listdistinguisher := function(L)
    local d, n, ld, i, k, m, newlist;

    d := [ ];
    n := Size(L);
    ld := [ ];
    for i in [1..n] do
        newlist := List(d, x->L[i][x]);
        if newlist in ld then
            k := Position(ld, newlist);
            m := PositionProperty([1..Size(L[i])], x->L[k][x] <> L[i][x]);
            Add(d, m);
            ld := List([1..i], j->List(d, x->L[j][x]));
        else
            Add(ld, newlist);
        fi;
    od;

    return rec(distinguisher := d, fingerprints := ld);
end;




# Example
# M24 on standard generators
# We wish to fingerprint (2,3,23)-pairs. This involves
# 2 2A/3B fingerprints, 2 2B/3A fingerprints and
# 1 2B/3B fingerprint.

M24example := function()
    local G, a, b, length, 2A, 2B, 3A, 3B,
          i, j, fpinfo, dist;
        
    # Standard generators of M24 taken from the Web Atlas
    a := (1,4)(2,7)(3,17)(5,13)(6,9)(8,15)(10,19)(11,18)
         (12,21)(14,16)(20,24)(22,23);
    b := (1,4,6)(2,21,14)(3,9,15)(5,18,10)(13,17,16)(19,24,23);
    G := Group(a, b);
    length := 100;
    
    2A := (x*y^2*(x*y)^3)^4;
    2B := x;
    3A := y;
    3B := (x*y^2*(x*y)^5)^7;
    
    fpinfo := fingerprintwords(G, a, b,
                      [[2A, 3B, 2, "2A/3B"],
                       [2B, 3A, 2, "2B/3A"],
                       [2B, 3B, 1, "2B/3B"]],
                      23,
                      length,
                      true);
    
    # Find a shorter fingerprint
    dist := listdistinguisher(Concatenation(fpinfo.fplist)).distinguisher;
    
    Print("Fingerprint: ",
          List(dist, x->fpinfo.words[x]), 
          "\n");
    
    for i in [1..Size(fpinfo.descriptions)] do
        Print(fpinfo.descriptions[i], "-----------\n");
        for j in [1..Size(fpinfo.conj[i])] do
            Print(fpinfo.conj[i][j], " - ");
            Print(List(dist, x->fpinfo.fplist[i][j][x]), "\n");
        od;
    od;
    
    return fpinfo;
    
end;


